"""Command line interface for rail-plot"""

from .project_commands import (
    project_cli,
    inspect_command,
    build_command,
    subsample_command,
    reduce_command,
    run_group,
)


__all__ = [
    "project_cli",
    "inspect_command",
    "build_command",
    "subsample_command",
    "reduce_command",
    "run_group",
]
