"""Command line interface for rail-plot"""

from .plot_commands import (
    plot_cli,
    run_command,
    inspect_command,
    extract_datasets_command,
    make_plot_groups_for_dataset_list,
    make_plot_groups_for_project,
)

__all__ = [
    "plot_cli",
    "run_command",
    "inspect_command",
    "extract_datasets_command",
    "make_plot_groups_for_dataset_list",
    "make_plot_groups_for_project",
]
