"""
The :py:mod:`rail.plotting` sub-package collects a set of tools to make
plots from the data generated by using the :py:mod:`rail.projects` tools.
"""

from . import control

__all__ = ["control"]
